import pandas as pd
from dataclasses import dataclass
from typing import Optional


REQUIRED_COLUMNS = {"type", "strike", "expiry", "price"}
OPTION_TYPES = {"C", "P"}


# ---------- Helpers ----------

def _normalize_option_types(series: pd.Series) -> pd.Series:
    """
    Normalize option type labels to C/P.
    Accepts: call, put, c, p, CE, PE, Call, Put, etc.
    """
    mapping = {
        "call": "C",
        "put": "P",
        "c": "C",
        "p": "P",
        "ce": "C",
        "pe": "P",
    }

    normalized = (
        series.astype(str)
        .str.strip()
        .str.lower()
        .map(mapping)
    )

    return normalized


@dataclass
class OptionChain:
    data: pd.DataFrame
    underlying: Optional[float] = None
    timestamp: Optional[pd.Timestamp] = None

    # ---------- Constructors ----------

    @classmethod
    def from_dataframe(cls, df: pd.DataFrame, underlying=None, timestamp=None):
        df = df.copy()

        # normalize option types BEFORE validation
        df["type"] = _normalize_option_types(df["type"])

        # normalize expiry to datetime
        df["expiry"] = pd.to_datetime(df["expiry"])

        cls._validate_schema(df)
        return cls(df, underlying, timestamp)

    @classmethod
    def from_csv(cls, path: str, **kwargs):
        df = pd.read_csv(path)
        return cls.from_dataframe(df, **kwargs)

    @classmethod
    def from_dict(cls, data: dict, **kwargs):
        df = pd.DataFrame(data)
        return cls.from_dataframe(df, **kwargs)

    @classmethod
    def from_nse(cls, json_data: dict):
        from qfinindia.data.nse import parse_nse_chain
        df, underlying, timestamp = parse_nse_chain(json_data)
        return cls.from_dataframe(df, underlying, timestamp)

    # ---------- Validation ----------

    @staticmethod
    def _validate_schema(df: pd.DataFrame):
        missing = REQUIRED_COLUMNS - set(df.columns)
        if missing:
            raise ValueError(f"Missing required columns: {missing}")

        if df["type"].isna().any():
            bad = df.loc[df["type"].isna(), "type"]
            raise ValueError(
                f"Invalid option type values found: {bad.unique()} "
                f"(allowed: call/put/CE/PE/c/p)"
            )

        if not set(df["type"].unique()).issubset(OPTION_TYPES):
            raise ValueError("Option type must normalize to C or P")

    # ---------- Filters ----------

    def calls(self):
        return OptionChain(
            self.data[self.data.type == "C"],
            self.underlying,
            self.timestamp
        )

    def puts(self):
        return OptionChain(
            self.data[self.data.type == "P"],
            self.underlying,
            self.timestamp
        )

    def expiry(self, expiry):
        expiry = pd.to_datetime(expiry)
        return OptionChain(
            self.data[self.data.expiry == expiry],
            self.underlying,
            self.timestamp
        )

    # ---------- Accessors ----------

    def strikes(self):
        return self.data["strike"].values

    def prices(self):
        return self.data["price"].values

    def iv(self):
        if "iv" not in self.data.columns:
            raise ValueError("IV column missing")
        return self.data["iv"].values

    def to_dataframe(self):
        return self.data.copy()



import numpy as np


class ScenarioAnalytics:
    """
    Market-implied scenario probabilities from risk-neutral density.
    """

    def __init__(self, strikes, density, spot):
        self.K = np.array(strikes)
        self.f = np.array(density)
        self.spot = spot

        # normalize
        self.f = self.f / np.trapz(self.f, self.K)

    def prob_below(self, level):
        mask = self.K <= level
        return np.trapz(self.f[mask], self.K[mask])

    def prob_above(self, level):
        mask = self.K >= level
        return np.trapz(self.f[mask], self.K[mask])

    def prob_between(self, low, high):
        mask = (self.K >= low) & (self.K <= high)
        return np.trapz(self.f[mask], self.K[mask])

    def expected_move_down(self):
        mask = self.K < self.spot
        prob = np.trapz(self.f[mask], self.K[mask])
        exp = np.trapz((self.spot - self.K[mask]) * self.f[mask], self.K[mask])
        return exp / prob if prob > 0 else 0

    def expected_move_up(self):
        mask = self.K > self.spot
        prob = np.trapz(self.f[mask], self.K[mask])
        exp = np.trapz((self.K[mask] - self.spot) * self.f[mask], self.K[mask])
        return exp / prob if prob > 0 else 0
    
    def tail_asymmetry(self):
        """
        Ratio of downside to upside probability mass.
        """
        down = self.prob_below(self.spot)
        up = self.prob_above(self.spot)
        return down / up if up > 0 else np.inf

    def expectation_asymmetry(self):
        """
        Ratio of expected down vs up move.
        """
        down = self.expected_move_down()
        up = self.expected_move_up()
        return down / up if up > 0 else np.inf




import numpy as np


class VolMetrics:

    def __init__(self, chain):
        self.chain = chain
        self.df = chain.to_dataframe()

    def forward(self):
        # mean of RND already best proxy
        from qfinindia.volatility.distribution import Distribution
        from qfinindia.volatility.rnd import RND

        expiry = self.df["expiry"].iloc[0]
        rnd = RND.from_chain(self.chain, expiry)
        dist = Distribution.from_rnd(rnd)
        return dist.mean()

    def atm_strike(self):
        fwd = self.forward()
        strikes = self.df["strike"].values
        return strikes[np.argmin(np.abs(strikes - fwd))]

    def atm_vol(self):
        K = self.atm_strike()
        row = self.df[(self.df["strike"] == K) & (self.df["type"] == "C")]
        if len(row) == 0:
            return np.nan
        return row["iv"].iloc[0]

    def skew(self):
        fwd = self.forward()
        strikes = self.df["strike"].values
        iv = self.df["iv"].values

        left = iv[strikes < fwd]
        right = iv[strikes > fwd]

        if len(left) == 0 or len(right) == 0:
            return np.nan

        return np.mean(left) - np.mean(right)



import numpy as np


class Distribution:

    def __init__(self, strikes, density):
        self.strikes = strikes
        self.density = density / np.trapz(density, strikes)

    @classmethod
    def from_rnd(cls, rnd):
        return cls(rnd.strikes, rnd.density)

    def mean(self):
        return np.trapz(self.strikes * self.density, self.strikes)

    def variance(self):
        m = self.mean()
        return np.trapz((self.strikes - m)**2 * self.density, self.strikes)

    def std(self):
        return np.sqrt(self.variance())

    def skew(self):
        m = self.mean()
        s = self.std()
        return np.trapz(((self.strikes - m)/s)**3 * self.density, self.strikes)

    def kurtosis(self):
        m = self.mean()
        s = self.std()
        return np.trapz(((self.strikes - m)/s)**4 * self.density, self.strikes)



from qfinindia.volatility.rnd import RND
from qfinindia.volatility.distribution import Distribution
from qfinindia.volatility.tailrisk import TailRisk
from qfinindia.volatility.vol_metrics import VolMetrics
import pandas as pd


def _compute_metrics(chain, expiry):
    """
    Internal: compute all implied parameters once
    """
    rnd = RND.from_chain(chain, expiry)
    dist = Distribution.from_rnd(rnd)
    tail = TailRisk.from_distribution(dist)
    vm = VolMetrics(chain)

    spot = chain.underlying
    fwd = dist.mean()
    move = dist.std()

    metrics = {
        "spot": float(spot),
        "forward": float(fwd),
        "forward_pct": float((fwd - spot) / spot),
        "expected_move": float(move),
        "atm_vol": float(vm.atm_vol()),
        "skew": float(dist.skew()),
        "var_5": float(tail.var(0.05)),
        "var_1": float(tail.var(0.01)),
        "bias": "Bullish" if fwd > spot else "Bearish" if fwd < spot else "Neutral",
    }

    return metrics


def generate_report(chain, expiry, output="text"):
    """
    Generate QFinIndia implied market report.

    Parameters
    ----------
    chain : OptionChain
    expiry : str or datetime
    output : {"text","dict","df"}

    Returns
    -------
    str | dict | pandas.DataFrame
    """
    m = _compute_metrics(chain, expiry)

    if output == "dict":
        return m

    if output == "df":
        return pd.DataFrame([m])

    # default text
    report = f"""
QFinIndia IMPLIED MARKET REPORT
--------------------------------
Spot: {m['spot']:.0f}
Forward: {m['forward']:.0f} ({m['forward_pct']*100:.2f}%)

Expected Move: ±{m['expected_move']:.0f}
ATM Vol: {m['atm_vol']:.2f}
Skew: {m['skew']:.2f}

VaR 5%: {m['var_5']:.0f}
VaR 1%: {m['var_1']:.0f}

Bias: {m['bias']}
"""
    return report.strip()



import numpy as np
from scipy.interpolate import CubicSpline


class RND:

    def __init__(self, strikes, density):
        self.strikes = strikes
        self.density = density

    @classmethod
    def from_chain(cls, chain, expiry, option_type="C"):
        sub = chain.expiry(expiry)

        if option_type == "C":
            sub = sub.calls()
        else:
            sub = sub.puts()

        df = sub.to_dataframe().sort_values("strike")

        K = df["strike"].values
        C = df["price"].values

        # spline price curve
        spline = CubicSpline(K, C)

        # second derivative (Breeden-Litzenberger)
        K_grid = np.linspace(K.min(), K.max(), 200)
        second_deriv = spline(K_grid, 2)

        density = np.maximum(second_deriv, 0)

        return cls(K_grid, density)

    def plot(self):
        import matplotlib.pyplot as plt
        plt.figure(figsize=(6,4))
        plt.plot(self.strikes, self.density)
        plt.xlabel("Strike")
        plt.ylabel("Density")
        plt.title("Risk-Neutral Density")
        plt.show()



class VolSurface:

    def __init__(self, df, underlying=None):
        self.df = df
        self.underlying = underlying

    @classmethod
    def from_chain(cls, chain):
        df = chain.to_dataframe().dropna(subset=["iv"])
        return cls(df, chain.underlying)



import numpy as np


class TailRisk:

    def __init__(self, strikes, density):
        self.strikes = strikes
        self.density = density / np.trapz(density, strikes)

    @classmethod
    def from_distribution(cls, dist):
        return cls(dist.strikes, dist.density)

    def var(self, alpha=0.05):
        cdf = np.cumsum(self.density)
        cdf /= cdf[-1]
        idx = np.searchsorted(cdf, alpha)
        return self.strikes[idx]

    def expected_shortfall(self, alpha=0.05):
        var_level = self.var(alpha)
        mask = self.strikes <= var_level
        prob = np.trapz(self.density[mask], self.strikes[mask])
        loss = np.trapz(self.strikes[mask] * self.density[mask], self.strikes[mask])
        return loss / prob if prob > 0 else np.nan

    def downside_probability(self, spot):
        mask = self.strikes < spot
        return np.trapz(self.density[mask], self.strikes[mask])

    def upside_probability(self, spot):
        mask = self.strikes >= spot
        return np.trapz(self.density[mask], self.strikes[mask])



import numpy as np
from scipy.stats import norm


class NSEOption:
    """
    Black-Scholes option pricing adapted for Indian markets.
    Provides price and Greeks.
    """

    def __init__(self, spot, strike, rate, vol, time):
        self.S = float(spot)
        self.K = float(strike)
        self.r = float(rate)
        self.sigma = float(vol)
        self.T = float(time)

    # ---------- core terms ----------
    def _d1(self):
        return (np.log(self.S / self.K) +
                (self.r + 0.5 * self.sigma**2) * self.T) / \
               (self.sigma * np.sqrt(self.T))

    def _d2(self):
        return self._d1() - self.sigma * np.sqrt(self.T)

    # ---------- pricing ----------
    def call_price(self):
        d1 = self._d1()
        d2 = self._d2()
        return self.S * norm.cdf(d1) - \
               self.K * np.exp(-self.r * self.T) * norm.cdf(d2)

    def put_price(self):
        d1 = self._d1()
        d2 = self._d2()
        return self.K * np.exp(-self.r * self.T) * norm.cdf(-d2) - \
               self.S * norm.cdf(-d1)

    # ---------- greeks ----------
    def call_delta(self):
        return norm.cdf(self._d1())

    def put_delta(self):
        return norm.cdf(self._d1()) - 1

    def gamma(self):
        return norm.pdf(self._d1()) / \
               (self.S * self.sigma * np.sqrt(self.T))

    def vega(self):
        return self.S * norm.pdf(self._d1()) * np.sqrt(self.T)

    def call_theta(self):
        d1 = self._d1()
        d2 = self._d2()
        term1 = -self.S * norm.pdf(d1) * self.sigma / (2 * np.sqrt(self.T))
        term2 = -self.r * self.K * np.exp(-self.r * self.T) * norm.cdf(d2)
        return term1 + term2

    def put_theta(self):
        d1 = self._d1()
        d2 = self._d2()
        term1 = -self.S * norm.pdf(d1) * self.sigma / (2 * np.sqrt(self.T))
        term2 = self.r * self.K * np.exp(-self.r * self.T) * norm.cdf(-d2)
        return term1 + term2
    
        # ---------- implied volatility ----------
    def implied_volatility(self, market_price, option_type="call",
                           tol=1e-6, max_iter=100):
        """
        Compute implied volatility using Newton-Raphson.
        """

        sigma = 0.2  # initial guess

        for _ in range(max_iter):
            self.sigma = sigma

            if option_type.lower() == "call":
                price = self.call_price()
            else:
                price = self.put_price()

            diff = price - market_price

            if abs(diff) < tol:
                return sigma

            vega = self.vega()

            if vega == 0:
                break

            sigma = sigma - diff / vega

        return sigma




import numpy as np


class SurfaceAnalytics:
    """
    Quantitative analytics on implied volatility surface.
    """

    def __init__(self, strikes, maturities, iv_surface, spot):
        self.strikes = np.array(strikes)
        self.maturities = np.array(maturities)
        self.iv = np.array(iv_surface)
        self.spot = spot

    # ---------- ATM term structure ----------
    def atm_term_structure(self):
        idx = np.argmin(np.abs(self.strikes - self.spot))
        return self.maturities, self.iv[:, idx]

    # ---------- skew slope ----------
    def skew(self):
        slopes = []
        for row in self.iv:
            slope = np.polyfit(self.strikes, row, 1)[0]
            slopes.append(slope)
        return self.maturities, np.array(slopes)

    # ---------- curvature ----------
    def curvature(self):
        curv = []
        for row in self.iv:
            quad = np.polyfit(self.strikes, row, 2)[0]
            curv.append(quad)
        return self.maturities, np.array(curv)



class MarketIntelligence:

    def __init__(self, chain, dist, tail):
        self.chain = chain
        self.dist = dist
        self.tail = tail

    @classmethod
    def from_metrics(cls, chain, dist, tail):
        return cls(chain, dist, tail)

    def forward(self):
        return self.dist.mean()

    def expected_move(self):
        return self.dist.std()

    def skew(self):
        return self.dist.skew()

    def var5(self):
        return self.tail.var(0.05)

    def var1(self):
        return self.tail.var(0.01)

    def bias(self):
        fwd = self.forward()
        spot = self.chain.underlying
        diff = (fwd - spot) / spot

        if diff > 0.002:
            return "Bullish"
        elif diff < -0.002:
            return "Bearish"
        else:
            return "Neutral"

    def summary(self):
        spot = self.chain.underlying
        fwd = self.forward()
        move = self.expected_move()

        return f"""
NIFTY IMPLIED OUTLOOK
---------------------
Spot: {spot:.0f}
Forward: {fwd:.0f} ({(fwd-spot)/spot*100:.2f}%)
Expected Move: ±{move:.0f}
Skew: {self.skew():.2f}

VaR 5%: {self.var5():.0f}
VaR 1%: {self.var1():.0f}

Bias: {self.bias()}
"""



import numpy as np
import matplotlib.pyplot as plt

class VolSmile:

    
    def __init__(self, strikes, iv, underlying=None, expiry=None):
        self.strikes = np.array(strikes)
        self.iv = np.array(iv)
        self.underlying = underlying
        self.expiry = expiry

    @classmethod
    def from_chain(cls, chain, expiry=None, option_type="C"):
        sub = chain

        if expiry is not None:
            sub = sub.expiry(expiry)

        if option_type == "C":
            sub = sub.calls()
        else:
            sub = sub.puts()

        df = sub.to_dataframe().dropna(subset=["iv"])

        return cls(
            strikes=df["strike"].values,
            iv=df["iv"].values,
            underlying=chain.underlying,
            expiry=expiry
        )
    
    
    def plot(self):
        plt.figure(figsize=(6,4))
        plt.plot(self.strikes, self.iv, marker="o")
        plt.xlabel("Strike")
        plt.ylabel("Implied Volatility")
        plt.title(f"Volatility Smile ({self.expiry})")
        plt.show()
